import 'dart:async';

import 'log_level.dart';
import 'log_record.dart';
import 'log_filter.dart';
import 'log_receiver.dart';

class Log {
  static List<ILogFilter> _filters = new List<ILogFilter>();
  static List<ILogReceiver> _receivers = new List<ILogReceiver>();

  static void registerFilter(ILogFilter filter) {
    _filters.add(filter);
  }

  static void unRegisterFilter(ILogFilter filter) {
    _filters.remove(filter);
  }

  static void clearFilter() {
    _filters.clear();
  }

  static void registerReceiver(ILogReceiver receiver) {
    _receivers.add(receiver);
  }

  static void unRegisterReceiver(ILogReceiver receiver) {
    _receivers.remove(receiver);
  }

  static void clearReceiver() {
    _receivers.clear();
  }

  static void _log(String module, LogLevel level, Object message,
      {Object exception, Zone zone, StackTrace stackTrace}) {

    // 组装LogRecord
    if (zone == null) {
      zone = Zone.current;
    }
    if (stackTrace == null) {
      stackTrace = StackTrace.current;
    }
    final LogRecord record = new LogRecord(module, level, message, exception, zone, stackTrace);

    final bool canAccept = _filteLog(record);
    if (canAccept) {
      _printLog(record);
    }
  }

  /// 输出日志
  static void _printLog(LogRecord record){
    if (_receivers.isEmpty){
      // 添加默认的控制台receiver
      registerReceiver(new _DefaultConsoleReceiver());
    }

    for (ILogReceiver receiver in _receivers) {
      receiver.onReceive(record);
    }
  }

  /// 过滤日志
  static bool _filteLog(LogRecord record){
    if (_filters.isEmpty){
      registerFilter(new _DefaultLogFilter());
    }

    bool canAccept = true;
    for (ILogFilter filter in _filters) {
      if (!filter.canAccept(record)) {
        canAccept = false;
        break;
      }
    }
    return canAccept;
  }

  static void trace(String module, Object message,
      {Object exception, Zone zone, StackTrace stackTrace}) {
    _log(module, LogLevel.trace, message,
        exception: exception, zone: zone, stackTrace: stackTrace);
  }

  static void debug(String module, Object message,
      {Object exception, Zone zone, StackTrace stackTrace}) {
    _log(module, LogLevel.debug, message,
        exception: exception, zone: zone, stackTrace: stackTrace);
  }

  static void info(String module, Object message,
      {Object exception, Zone zone, StackTrace stackTrace}) {
    _log(module, LogLevel.info, message,
        exception: exception, zone: zone, stackTrace: stackTrace);
  }

  static void warning(String module, Object message,
      {Object exception, Zone zone, StackTrace stackTrace}) {
    _log(module, LogLevel.warning, message,
        exception: exception, zone: zone, stackTrace: stackTrace);
  }

  static void error(String module, Object message,
      {Object exception, Zone zone, StackTrace stackTrace}) {
    _log(module, LogLevel.error, message,
        exception: exception, zone: zone, stackTrace: stackTrace);
  }

  static void fatal(String module, Object message,
      {Object exception, Zone zone, StackTrace stackTrace}) {
    _log(module, LogLevel.fatal, message,
        exception: exception, zone: zone, stackTrace: stackTrace);
  }

}

/// 生产环境只输出warning及以上级别的log
class _DefaultLogFilter implements ILogFilter {

  @override
  bool canAccept(LogRecord record) {
    const bool inProduction = const bool.fromEnvironment("dart.vm.product");
    if (inProduction) {
      if (record.level == LogLevel.warning ||
          record.level == LogLevel.error ||
          record.level == LogLevel.fatal ||
          record.level == LogLevel.crash) {
        return true;
      }else{
        return false;
      }
    } else {
      return true;
    }
  }
}

/// 默认的控制台Receiver
class _DefaultConsoleReceiver implements ILogReceiver{

  @override
  void onReceive(LogRecord record){
    print(record.format());
  }

}
